package com.unclouded.android;

import java.net.InetAddress;


/**
 *  Monitors the state of the established network connection.
 *  
 *  A NetworkListener instance is triggered in four cases:
 *  <ul>
 *  	<li>{@link #isOnline(InetAddress)} is triggered after 
 *  	the device is online and obtained a network address.</li>
 *  	<li>{@link #isOffline(InetAddress)} s triggered after 
 *  	the device is offline from the network. </li>
 *  	<li>{@link #isConnectedTo(InetAddress)} is triggered after 
 *  	the device connects to another device in the network.</li>
 *  	<li>{@link #isDisconnectedFrom(InetAddress)} is triggered after 
 *  	some device in the network has disconnected from the network.</li>
 *  </ul>
 *  
 *  <p>Note that all four methods are optional and not required to be implemented.</p>
 *
 */
public class NetworkListener {
	
	/**
	 * Triggered after the device is online.
	 * The device is now connected to the network and can be
	 * discovered by other devices.
	 * 
	 * @param ipAddress  the address that was obtained upon connecting to the network. 
	 */
	public void isOnline(InetAddress ipAddress){
		return;
	}
	
	
	/**
	 * Triggered after the device is offline.
	 * The device is now disconnected from the network and cannot be
	 * discovered by other devices, nor interact with them. 
	 * 
	 * @param ipAddress  the address that was obtained when the device was online.
	 */
	public void isOffline(InetAddress ipAddress){
		return;
	}
	
	
	/**
	 * Triggered after establishing a connection to a device in the network.
	 * 
	 * @param ipAddress  the address of the devices wherewith a connection is established.
	 */
	public void isConnectedTo(InetAddress ipAddress){
		return;
	}
	
	
	/**
	 * Triggered after a disconnection of a device in the network.
	 * 
	 * @param ipAddress the address of the device that disconnected from the network.
	 */
	public void isDisconnectedFrom(InetAddress ipAddress){
		return;
	}
	
	
	/**
	 * Returns a string representation of a NetworkListener object. 
	 * 
	 * @return a string representation of the network listener.
	 */
	public String toString(){
		return "[NetworkListener]";
	}
	
	

}
