package com.unclouded.android;

/**
 * Monitors the state of one or more services.
 * 
 * <p>A {@link ServiceListener} is passed as argument to the
 * {@link Unclouded#when(TypeTag, ServiceListener) when} or 
 * {@link Unclouded#whenever(TypeTag, ServiceListener) whenever} method on 
 * the {@link Unclouded} event loop. It triggers notifications in three cases:
 * 
 * <ul>
 * 	<li>{@link #isDiscovered(Object)} is triggered 
 * 		when the service is discovered for the first time.</li>
 * 	<li>{@link #isDisconnected(Object)} is triggered 
 * 		when the service is disconnected from the network.</li>
 * 	<li>{@link #isReconnected(Object)} is triggered 
 * 		when the service is reconnected to the network.</li>
 * </ul>
 * </p>
 * 
 * <p>Note that only {@link #isDiscovered(Object)} is required to be implemented.
 * Both {@link #isDisconnected(Object)} and {@link #isReconnected(Object)} are optional.</p>
 *
 * @param <T> the type of data that is expected when discovering a service.
 */
public abstract class ServiceListener<T> {
	
	/**
	 * Triggered when the service is discovered in the network for the first time.
	 * @param value  the value or {@link RemoteReference} hold by the service.
	 */
	public abstract void isDiscovered(T value);
	
	/**
	 * Triggered when the service is disconnected from the network.
	 * @param value  the value or {@link RemoteReference} hold by the service.
	 */
	public void isDisconnected(T value){ return; }
	
	
	/**
	 * Triggered when the service is reconnected to the network.
	 * @param value  the value or {@link RemoteReference} hold by the service.
	 */
	public void isReconnected(T value){ return; }

}
