package com.unclouded.android;

import java.io.Serializable;

import com.unclouded.android.internal.InternalEventLoop;


/**
 * Represents the Unclouded event loop. 
 * This is the central entity that manages all communication and 
 * lets you perform fundamental Unclouded tasks, such as establishing 
 * a connection to the {@link Network}, publishing a service and 
 * discovering services of a particular {@link TypeTag}.
 * 
 * <p>The {@link Unclouded} event loop adopts event-driven communication and 
 * consists of an event queue that stores (incoming) events. 
 * These events are sequentially processed, one at a time.
 * Interactions between event loops are processed asynchronously.</p>
 * 
 * <p>To get the {@link Unclouded} instance representing the local Unclouded
 * event loop, call the static {@link #getInstance} method.</p>
 * 
 * <b>Developer Guide</b><br>
 * For more information about using Unclouded, read the
 * <a href="http://localhost:3000/docs/android" target="_blank">Unclouded</a> developer guide.
 */

public class Unclouded {
	
	private static Unclouded instance;
	private InternalEventLoop eventLoop;
	
	
	/**
     * Get a handle to the Unclouded instance.
     * @return the Unclouded instance.
     */
	static public Unclouded getInstance(){
		if(instance == null){
			instance = new Unclouded();
		}
		return instance;
	}
	
	private Unclouded(){
		eventLoop = InternalEventLoop.getInstance();
	}
	
	/**
     * Connect to the network and obtain a network address.
     * @return An {@link Network} object to listen for network status updates.
     */
	public Network goOnline() {
		return eventLoop.goOnline();
	}
	
	/**
     * Disconnect from the network.
     */
	public void goOffline() {
		eventLoop.goOffline();
	}
	
	
	/**
     * Disconnect from the network and reset the event loop.
     */
	public void reset(){
		eventLoop.reset();
	}
	
	// --------------------------------------------------
	/**
     * Broadcast a service referencing an {@link UObject} across the network.
     * <p>An object implementing the {@link UObject} interface is passed 
     * by reference. When received by the recipient, it is referenced by an {@link RemoteReference}.</p>
     * 
     * @param typeTag  type of the service to be broadcasted.
     * @param obj  any object implementing the {@link UObject} interface.
     * @return A {@link ServicePublication} instance to manage the publication.
     */
	public ServicePublication broadcast(TypeTag typeTag, UObject obj) {
		return eventLoop.broadcast(typeTag, obj);
	}
	
	/**
     * Broadcast a service holding a {@link String} across the network.
     * 
     * @param typeTag  type of the service to be broadcasted.
     * @param s  the String to broadcast.
     * @return A {@link ServicePublication} instance to manage the publication.
     */
	public ServicePublication broadcast(TypeTag typeTag, String s) {
		return eventLoop.broadcast(typeTag, s);
	}
	
	/**
     * Broadcast a service holding a {@link Integer} across the network.
     * 
     * @param typeTag  type of the service to be broadcasted.
     * @param i  the Integer to broadcast.
     * @return A {@link ServicePublication} instance to manage the publication.
     */
	public ServicePublication broadcast(TypeTag typeTag, Integer i) {
		return eventLoop.broadcast(typeTag, i);
	}
	
	/**
     * Broadcast a service holding a {@link Double} across the network.
     * 
     * @param typeTag  type of the service to be broadcasted.
     * @param d  the Double to broadcast.
     * @return A {@link ServicePublication} instance to manage the publication.
     */
	public ServicePublication broadcast(TypeTag typeTag, Double d) {
		return eventLoop.broadcast(typeTag, d);
	}
	
	/**
     * Broadcast a service holding a {@link Serializable} object across the network.
     * 
     * @param typeTag  type of the service to be broadcasted.
     * @param serializable  the serializable object to broadcast.
     * @return A {@link ServicePublication} instance to manage the publication.
     */
	public ServicePublication broadcast(TypeTag typeTag, Serializable serializable){
		return eventLoop.broadcast(typeTag, serializable);
	}
	
	
	// --------------------------------------------------
	/**
     * Listen for a single service of a particular {@link TypeTag} in the network.
     * 
     * @param typeTag  type of the service to be discovered.
     * @param listener  used to trigger callback methods when status of the discovered service in the network changes.
     * @return A {@link ServiceSubscription} instance to manage the subscription.
     */
	public ServiceSubscription when(TypeTag typeTag, ServiceListener<? extends Object> listener){
		return eventLoop.when(typeTag, listener);
	}
	
	
	/**
     * Listen for all services of a particular {@link TypeTag} in the network.
     * 
     * @param typeTag  type of the service to be discovered.
     * @param listener  used to trigger callback methods when status of the discovered services in the network change.
     * @return A {@link ServiceSubscription} instance to manage the subscription.
     */
	public ServiceSubscription whenever(TypeTag typeTag, ServiceListener<? extends Object> listener){
		return eventLoop.whenever(typeTag, listener);
	}
	

}
