/**
 * AmbientTalk/2 Project
 * CommandProcessor.java created on 3-apr-2007 at 15:02:51
 * (c) Programming Technology Lab, 2006 - 2007
 * Authors: Tom Van Cutsem & Stijn Mostinckx
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */
package com.unclouded.android.comm;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.Socket;

import com.unclouded.android.cmd.Command;
import com.unclouded.android.logging.Logging;

/**
 * A command processor is responsible for:
 *  <ul>
 *   <li>Reading and executing incoming Actor Command objects from an incoming socket connection
 *   <li>Signaling to the communication bus that the connection has died upon a network exception
 *  </ul>
 * 
 * From the moment the incoming connection raises any kind of exception, the command processor
 * is stopped and the communication bus is notified that the connected slave should be removed
 * from the connections table. At that point, the thread's body will stop executing, which
 * should cause the command processor to become subject to garbage collection (because the
 * {@link MasterConnectionThread} does not store any reference to the created processor).
 * 
 * @author tvcutsem
 */
public class CommandProcessor extends Thread {

	private final Address remoteActorAddress;
	private final Socket socket;
	private final ObjectInputStream inputStream;
	private final CommunicationBus communicationBus;
	
	public CommandProcessor(Address remoteActorAddress_, Socket socket_, ObjectInputStream inputStream_, CommunicationBus bus_) throws IOException {
		super("CommandProcessor for " + remoteActorAddress_);
		remoteActorAddress = remoteActorAddress_;
		socket = socket_;
		inputStream = inputStream_;
		communicationBus = bus_;
	}
	
	/**
	 * Reads {@link Command} objects from the incoming socket connection and executes them.
	 * The thread does this eternally until network connection is disrupted, or the socket is
	 * closed by the communication bus.
	 * 
	 * Note: it is the responsibility of this thread to properly unregister and remove the
	 * connection from the communication bus when an I/O error of any kind has occured.
	 */
	public void run() {
		try {
			while (true) {
				Command command = (Command) inputStream.readObject();
				Logging.CommandProcessor_LOG.debug("New command "+command+" received from "+remoteActorAddress);
				command.uponReceiptBy(communicationBus.host, remoteActorAddress);
			}
		} catch(Exception e) {
		} finally {
			Logging.CommandProcessor_LOG.debug("Something happened to "+remoteActorAddress+" . Connection being removed.");
			communicationBus.removeConnection(remoteActorAddress, socket);
		}
	}
	
	public String toString() {
		return "CommandProcessorFor["+remoteActorAddress+"]";
	}

}
