package com.unclouded.android.comm;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Timer;
import java.util.TimerTask;

import com.unclouded.android.cmd.CMDInitializeConnection;
import com.unclouded.android.cmd.Command;
import com.unclouded.android.internal.InternalEventLoop;
import com.unclouded.android.logging.Logging;


public class CommunicationBus {
	
	static public enum STATUS { NEW, REGISTERED, CLOSED, TIMEDOUT };
	
	/** the actor that owns this communication bus */
	public final InternalEventLoop host;
	
	/** the name of the overlay network in which to discover actors */
	static private final String groupName = "Unclouded";
	
	/** the ip address to which connect */
	static private final String ipAddress = "0.0.0.0";
	
	/** if non-null, the communication bus is connected to the network */
	private volatile Address networkAddress;
	
	/** The timer used for checking 'stale' (i.e. timed out) connections in the connection table. */
	private final Timer timeoutDetectorTimer;
	
	/** The timer task used for removing timed out connections from the connections table. */
	private TimeoutDetectorTask timeoutDetector;
	
	/** Maps the address of a currently connected actor to connection information
	 *  such as the last time it was seen, and the socket connection.
	 *  Also known as the "Connection table". Access to it must be synchronized. */
	private final HashMap<Address, Connection> addressToConnection;
	
	/** Listen for incomming slave connections */
	private MasterConnectionThread masterConnectionThread;
	
	/** Listen for the heartbeats of other actors */
	private MulticastListenerThread connectionListener;
	
	/** Broadcast the heartbeat at a fixed rate */
	private MulticastServerThread connectionServer;
	
	
	
	/** -----------------------------------------------
	 * 	CONSTRUCTOR
	 *  ----------------------------------------------- */
	
	public CommunicationBus(InternalEventLoop host_){
		host = host_;
		timeoutDetectorTimer = new Timer(true); // create a daemon timer
		addressToConnection = new HashMap<Address, Connection>();
	}
	
	
	
	/** -----------------------------------------------
	 * 	GETTER METHODS
	 *  ----------------------------------------------- */
	
	static public String getIpAddress() { return ipAddress; }
	
	static public String getGroupName(){ return groupName; }
	
	public boolean isOnline(){ return (networkAddress != null); }
	
	public Address getNetworkAddress(){ return networkAddress; }
	
	/** -----------------------------------------------
	 * 
	 * CONNECT
	 * 
	 * Tries to connect the communication bus to the underlying network.
	 * @throws IOException if no server socket could be created to listen
	 * for incoming connections. If this exception is raised, it is
	 * guaranteed that the communication bus is left disconnected (i.e.
	 * it is not partially connected)
	 * 
	 *  ----------------------------------------------- */
	
	public Address connect() {
		if (isOnline()) {
			return networkAddress; // if the bus is already connected, there is no need to connect it again
		}
		masterConnectionThread = new MasterConnectionThread(this);
		try {
			networkAddress = masterConnectionThread.startServing();
		} catch (IOException e) {
			masterConnectionThread = null;
		}
		connectionListener = new MulticastListenerThread(this, networkAddress);
		connectionServer = new MulticastServerThread(networkAddress);
		connectionListener.start();
		connectionServer.start();
		
		// start detecting timed out actors
		timeoutDetector = new TimeoutDetectorTask();
		timeoutDetectorTimer.scheduleAtFixedRate(timeoutDetector, 0, TimeoutDetectorTask.DETECTION_RATE);
		return networkAddress;
	}
	
	
	
	/** -----------------------------------------------
	 * 
	 * DISCONNECT
	 * 
     * Called by the VM when it has disconnected from the underlying channel.
     * It gracefully shuts down all network threads, sets the network address
     * to null and removes all current connections from the connection table.
     *
     *  ----------------------------------------------- */
	
	public void disconnect() {
		if (!isOnline()) {
			return; // if the bus is already disconnected, there is no need to take it offline again
		}
		// once the bus is disconnected, the network address is set to null.
		// this ensures that no further incoming connections are allowed to be
		// registered in the addConnection method!
		networkAddress = null;
		masterConnectionThread.stopServing();
		connectionListener.stopListening();
		connectionServer.stopBroadcasting();
		masterConnectionThread = null;
		connectionListener = null;
		connectionServer = null;
		
		// stop detecting timed out VMs
		timeoutDetector.cancel();
		timeoutDetector = null;
		
		closeConnectionOfAllMembers();
		
		synchronized(addressToConnection){
			addressToConnection.clear();
		}
	}
	
	
	
	/** -----------------------------------------------
	 * 
	 * ADD NEW CONNECTION
	 * 
	 * Registers a new virtual machine connection for the given address.
	 * If this Actor was a slave in the discovery process, this method is
	 * invoked by the {@link MulticastListenerThread}. If this Actor was
	 * a master in the discovery process, this method is invoked by the
	 * {@link MasterConnectionThread}.
	 * 
	 * The socket's output stream will be stored in the connection table and is used
	 * for transmitting Actor Commands to this Actor. The socket's input stream will be
	 * coupled to a dedicated {@link CommandProcessor} which is responsible for
	 * processing incoming Actor command objects.
	 * 
	 * Calling this method implicitly also triggers a memberJoined event on this actor
	 * 
	 * @throws IOException if no ObjectOutputStream can be created for the socket's output stream,
	 * or if no ObjectInputStream can be created for the socket's input stream.
	 * If this exception is raised, it is guaranteed the member is not registered in the connection table.
	 * 
	 *  ----------------------------------------------- */
	
	protected void addConnection(Address newMember, Socket socket) throws IOException {
		// if the bus has been disconnected, do not accept any new connections
		if (!isOnline()) { return; }
		// create a new connection object that can be used to send command objects to this member
		Connection registeredConnection = new Connection(socket, System.currentTimeMillis());
		// spawn a new command processor dedicated for handling the command objects received from this member
		CommandProcessor processor = new CommandProcessor(newMember, socket, registeredConnection.incoming_, this);
		synchronized (addressToConnection) {
			// first check whether a connection for this member already exists
			// (may happen when an old connection for this member has not yet been deleted)
			// This case IS possible when a master receives a new connection from a slave,
			// but that master has not yet detected the disconnection of the slave's old connection
			Connection oldConnection = addressToConnection.get(newMember);
			if (oldConnection != null) {
				// explicitly close the connection we're about to override
				// Note: in this case, we should *not* signal to the VM that a new member joined,
				// because in the corresponding removeConnection, we will also not signal a member left
				// (i.e. the connection was restored quickly enough such that we can abstract over the disconnection)
				oldConnection.close();
			} 
			//Logging.CommunicationBus_LOG.debug("Add connection to table");
			// register the new connection in the table (this may override an old connection)
			addressToConnection.put(newMember, registeredConnection);
			// Upon observing this actor for the first time, give handshake
			if (oldConnection == null) {
				Logging.CommunicationBus_LOG.debug("New connection discovered");
				// Send handshake 
				new CMDInitializeConnection(host.getId()).send(this, newMember);
			}
		}
		processor.start();
	}
	
	
	
	/** ------------------------------------------------
	 * 
	 * REMOVE CONNECTION
	 * 
	 * It is the responsibility of the {@link CommandProcessor} tied to the given Actor
	 * to invoke this method when its connection has failed.
	 * 
	 * This is the <b>only</b> method responsible for removing entries from the connection table.
	 * Calling this method implicitly also triggers a memberLeft event on this VM
	 * 
	 *  ----------------------------------------------- */
	
	protected void removeConnection(Address oldMember, Socket connSocket) {
		synchronized (addressToConnection) {
			Connection conn = addressToConnection.get(oldMember);
			if (conn != null) {
				// ONLY delete the connection if that connection was registered for the given
				// socket. It might be that the connection in the connection table is already
				// a NEWER connection that has OVERWRITTEN the old one.
				if (conn.socket_ == connSocket) {
					if(conn.close() == STATUS.TIMEDOUT){
						addressToConnection.remove(oldMember);
					}
					host.event_memberLeft(oldMember);
				}
				// Note: if the sockets do not match, then the connection to be
				// removed by this call has already been replaced by a more recent connection to the same address
				// See the code for addConnection for more details
				// Because the connection to be removed was already replaced, it is not necessary
				// to notify the VM that a member has left
			}
		}
	}
	
	
	
	/** -----------------------------------------------
	 * 
	 * OBSERVE TIMEOUT
	 * 
	 * The {@link MulticastListenerThread} invokes this method regularly to
	 * remove connections to VMs which have not responded for longer than the
	 * given timeout period.
	 * 
	 * Removal is done implicitly by closing timed out connections. The
	 * {@link CommandProcessor} tied to each connection is responsible for
	 * actually removing it. Calling this method might also implicitly trigger 
	 * one or more memberLeft events on this VM.
	 * 
	 *  ----------------------------------------------- */
	
	protected void closeConnectionOfMembersNotSeenIn(long period) {
		synchronized (addressToConnection) {
			long now = System.currentTimeMillis();
			Collection<Connection> connections = addressToConnection.values();
			for (Iterator<Connection> iter = connections.iterator(); iter.hasNext();) {
				Connection c = iter.next();
				//Address address = entry.getKey();
				if (now - c.lastSeenAtTime_ > period) {
					boolean socketBeingClosed = c.timedOut(); // If yes, the entry will be deleted by the Command Processor
					if(!socketBeingClosed){ // If false, socket was already closed (by disconnection), so remove ourselves
						iter.remove();
					}
				}
			}
		}
	}
	
	
	/** -----------------------------------------------
	 * 
	 * OBSERVE TIMEOUT
	 * 
	 * When the communication bus is explicitly disconnected from the
	 * network, the actor should be notified that all connected actors have disconnected.
	 * 
	 * Removal is done implicitly by closing all connections. The
	 * {@link CommandProcessor} tied to each connection is responsible for
	 * actually removing them.
	 * 
	 *  ----------------------------------------------- */
	
	private void closeConnectionOfAllMembers() {
		synchronized (addressToConnection) {
			Collection<Connection> connections = addressToConnection.values();
			for (Iterator<Connection> iter = connections.iterator(); iter.hasNext();) {
				Connection c = iter.next();
				c.close(); // ensures the command processor will remove the entry eventually
			}
		}
	}
	
	
	
	/** -----------------------------------------------
	 * 
	 * UPDATE TIMESTAMP
	 * 
	 * Updates the time that the given actor was last seen online.
	 * @param member the address of the detected actor
	 * @return true if the actor was properly registered, false if the actor is not yet registered
	 * 
	 *  ----------------------------------------------- */
	
	protected STATUS updateTimeLastSeen(Address member) {
		synchronized (addressToConnection) {
			Connection conn = addressToConnection.get(member);
			if (conn != null) {
				if(conn.updateTimestamp(System.currentTimeMillis())){
					return STATUS.REGISTERED;
				} else {
					return STATUS.CLOSED;
				}
			} else {
				return STATUS.NEW;
			}
		}
	}
	
	
	
	/** -----------------------------------------------
	 * 
	 * SEND ASYNCHRONOUS MULTICAST
	 * 
	 * Sends a VM Command object asynchronously to all connected VMs.
	 * There are no delivery guarantees for this message, nor is it guaranteed
	 * that all currently connected VMs will receive the message
	 * 
	 *  ----------------------------------------------- */
	
	@SuppressWarnings("unchecked")
	public void sendAsyncMulticast(Command msg) {
		// first clone the connection table such that we do not need to acquire the
		// lock for the entire duration of the multicast
		HashMap<Address, Connection> clonedConnections;
		synchronized (addressToConnection) {
			clonedConnections  = (HashMap<Address, Connection>) addressToConnection.clone();
		}
		for (Iterator<Entry<Address, Connection>> iter = clonedConnections.entrySet().iterator(); iter.hasNext();) {
			Map.Entry<Address, Connection> entry = (Map.Entry<Address, Connection>) iter.next();
			Address recipient = entry.getKey();
			Connection conn = entry.getValue();
			sendOneAsyncMessage(conn, msg, recipient);
		}
	}
	
	
	/** -----------------------------------------------
	 * 
	 * SEND ASYNCHRONOUS UNICAST
	 * 
	 * Sends a Actor Command object asynchronously to the recipient actor.
	 * There are no delivery guarantees for this message.
	 * If the recipient is offline, or the message times out, it is simply discarded
	 * 
	 *  ----------------------------------------------- */
	
	
	public void sendAsyncUnicast(Command msg, Address recipientActor) {
		Connection conn;
		synchronized (addressToConnection) {
			conn = addressToConnection.get(recipientActor);
		}
		if (conn != null) {
			sendOneAsyncMessage(conn, msg, recipientActor);
		} else {
			Logging.ERROR_LOG.debug("No connection found upon sending async unicast");
		}
	}
	
	
	/** -----------------------------------------------
	 * 
	 * SEND PARTICULAR ASYNCHRONOUS MESSAGE
	 * 
	 * Note that an I/O exception during the transmission of a command object is <b>fatal</b>
	 * for the connection. The connection is immediately terminated by closing the socket.
	 * It is the responsibility of the {@link CommandProcessor} registered on the socket
	 * connection's input stream to remove the recipient actor from the connection table.
	 * 
	 *  ----------------------------------------------- */
	
	private void sendOneAsyncMessage(Connection conn, Command msg, Address recipientActor) {
		try {
			conn.send(msg);
		} catch (IOException e) {
			// it is the sender's responsibility to close the connection's socket if something goes wrong
			// the CommandProcessor will remove the member from the connection table
			Logging.ERROR_LOG.debug("Error upon sending async msg. Connection terminated.");
			conn.close();
		}
	}
	
	
	
	/** -----------------------------------------------
	 * 
	 * SEND SYNCHRONOUS UNICAST MESSAGE
	 * 
	 * Sends the given {@link Command} to the given member. With 'synchronous' transmission,
	 * it is meant that if this method returns gracefully, the caller can rest assured that
	 * the remote actor has correctly received the command object. It does not given any guarantees
	 * that the command object will have been executed remotely.
	 *
	 *  ----------------------------------------------- */
	
	public void sendSynchronousUnicast(Command msg, Address recipientActor) {
		Connection conn;
		synchronized (addressToConnection) {
			conn = (Connection) addressToConnection.get(recipientActor);
		}
		if (conn == null) { return; } // Recipient actor offline	
		try {
			conn.send(msg);
		} catch (IOException e) {
			// it is the sender's responsibility to close the connection's socket if something goes wrong
			// the corresponding CommandProcessor registered on this socket will remove the member from
			// the connection table
			Logging.ERROR_LOG.debug("Error upon sending sync unicast. Connection terminated.");
			conn.close();
		} 
	}
	
	
	
	/** -----------------------------------------------
	 * 
	 * CONNECTION CLASS
	 * 
	 * Bookkeeping datastructure to store connection-related information
	 * of a connected VM. A connection entry is uniquely identified by
	 * means of the {@link Address} of its VM.
	 * 
	 *  ----------------------------------------------- */
	

	private static class Connection {
		public final Socket socket_;
		/** this value is updated as new heartbeats are received */
		public long lastSeenAtTime_;
		public final ObjectOutputStream outgoing_;
		public final ObjectInputStream incoming_;
		private STATUS status;
		public Connection(Socket s, long lastSeenAt) throws IOException {
			socket_ = s;
			status = STATUS.REGISTERED;
			lastSeenAtTime_ = lastSeenAt;
			// NOTE: apparently it is highly important that the ObjectOutputStream on the socket.getOutputStream() is
			// created BEFORE trying to create an ObjectInputStream on the socket.getInputStream()
			// switching the below two statements causes the master and the slave to deadlock!
			outgoing_ = new ObjectOutputStream(new BufferedOutputStream(s.getOutputStream()));
			// The buffered output stream must be EXPLICITLY flushed, otherwise the buffered input stream
			// that is created below (but by the other VM) will block indefinitely
			outgoing_.flush();
			incoming_ = new ObjectInputStream(new BufferedInputStream(s.getInputStream()));
		}
		
		public synchronized boolean updateTimestamp(long timeStamp){
			if(status == STATUS.CLOSED){ return false; }
			lastSeenAtTime_ = timeStamp;
			return true;
		}
		
		/**
		 * Closes the underlying socket. This will eventually trigger the command processor
		 * tied to this connection, which will cause this connection to be removed from the table.
		 */
		public synchronized STATUS close() {
			if(status != STATUS.TIMEDOUT && status != STATUS.CLOSED){
				try {
					socket_.close();
					status = STATUS.CLOSED;
				} catch (IOException e) { }
			}
			return status;
		}
		
		public synchronized boolean timedOut() {
			boolean socketClosed = false;
			if(status != STATUS.TIMEDOUT && status != STATUS.CLOSED){
				try {
					socket_.close();
					socketClosed = true;
				} catch (IOException e) { }
			}
			status = STATUS.TIMEDOUT;
			return socketClosed;
		}
		
		
		public synchronized void send(Command msg) throws IOException{
			outgoing_.writeObject(msg);
			outgoing_.flush();
		}
	}
	
	
	
	/** -----------------------------------------------
	 * 
	 * TIMEOUT DETECTOR
	 * 
	 * The timer task used for removing timed out connections
	 * from the connection table.
	 * 
	 *  ----------------------------------------------- */
	

	private class TimeoutDetectorTask extends TimerTask {
		/**
		 * the maximum amount of time that a remote VM gets to send a new
		 * heartbeat before it is considered as being 'offline'
		 */
		private static final int MAX_RESPONSE_DELAY = 10000; // in milliseconds
		
		/** The rate at which to schedule this timer task */
		public static final int DETECTION_RATE = 4000; // in milliseconds
		
		public void run() {
			closeConnectionOfMembersNotSeenIn(MAX_RESPONSE_DELAY);
		}
	}

}
