package com.unclouded.android.eventloop;

import com.unclouded.android.logging.Logging;


public class EventLoop {
	
	/**
	 * The event loop's event queue is a synchronized queue of Event objects.
	 * It is the sole communication channel between different event loops.
	 * As such, it is the means by which different actors - and event actors and
	 * their virtual machines - communicate.
	 */
	protected final EventQueue eventQueue_;
	
	/**
	 * Each event loop has an event processor, which is a thread responsible
	 * for perpetually dequeuing events from the event queue and passing them
	 * on to this event loop's handle method.
	 */
	protected Thread processor_;
	
	protected volatile boolean askedToStop_;
	
	
	/** -----------------------------------------------
	 * CONSTRUCTOR
	 * Note: the creator must explicitly call {@link this#start()} to start the event loop!
	 * ----------------------------------------------- */
	
	public EventLoop() {
		eventQueue_ = new EventQueue();
		askedToStop_ = false;
		processor_ = new EventProcessor();
	}
	
	
	public String toString() { return "EventLoop"; }
	
	protected final EventLoop owner() { return this; }
	
	
	/** -----------------------------------------------
	 * 	START EVENT LOOP
	 *  ----------------------------------------------- */
	
	public void start() {
		if(!processor_.isAlive()){
			Logging.Eventloop_LOG.debug("Eventloop started");
			processor_.start();
		}
	}
	
	
	/** -----------------------------------------------
	 * 	STOP EVENT LOOP
	 *  ----------------------------------------------- */
	
	public void stop() {
		if(processor_.isAlive()){
			Logging.Eventloop_LOG.debug("Eventloop stoped");
			askedToStop_ = true;
			// explicitly interrupt event processor because it may be blocked waiting on other events
			processor_.interrupt();
			// Remove all remaining events
			eventQueue_.flush();
		}
	}
	
	
	/** -----------------------------------------------
	 * RECEIVE EVENTS
	 * When an event loop receives an asynchronously emitted event, this message is
	 * immediately placed into its incoming event queue and will be processed later.
	 *  ----------------------------------------------- */
	
	protected final void receive(Event event) {
		eventQueue_.enqueue(event);
	}
	
	
	/** -----------------------------------------------
	 * 	HANDLE EVENTS
	 *  Handle events from the event queue.
	 *  ----------------------------------------------- */
	
	public void handle(Event event){
		event.process(this); // make the event process itself
	}
	
	
	
	/** -----------------------------------------------
	 * EXECUTE
	 * Invoked by the event processor thread continually while the event loop is alive.
	 * Continually dequeue events and make a subclass handle the event.
	 *  ----------------------------------------------- */
	
	protected void execute() {
		try {
			Event event = eventQueue_.dequeue();
			Logging.Eventloop_LOG.debug("Processing event: " + event);
			handle(event);
		} catch (InterruptedException e) { }
	}
	

	
	
	/** -----------------------------------------------
	 * 
	 * EVENT PROCESSOR
	 * 
	 * EventProcessor is a thread subclass whose primary goal is to keep a reference to the 
	 * associated event loop, which is used to transform threads into the more manageable 
	 * event loops
	 * 
	 * ----------------------------------------------- */
	
	public final class EventProcessor extends Thread {
		
		public final void run() {
			while(!askedToStop_) {
				execute();
				Thread.yield(); // give other event loops a chance to process an event
			}
		}
        
	}

}
