package com.unclouded.android.internal;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.LinkedHashMap;

import com.unclouded.android.Network;
import com.unclouded.android.ServiceListener;
import com.unclouded.android.ServicePublication;
import com.unclouded.android.ServiceSubscription;
import com.unclouded.android.TypeTag;
import com.unclouded.android.UObject;
import com.unclouded.android.cmd.CMDCancelService;
import com.unclouded.android.cmd.CMDTransmitLetter;
import com.unclouded.android.cmd.CMDTransmitLetterReply;
import com.unclouded.android.comm.ActorAddressBook;
import com.unclouded.android.comm.Address;
import com.unclouded.android.comm.CommunicationBus;
import com.unclouded.android.eventloop.Event;
import com.unclouded.android.eventloop.EventLoop;
import com.unclouded.android.ids.EventLoopId;
import com.unclouded.android.ids.PromiseId;
import com.unclouded.android.ids.PublicationId;
import com.unclouded.android.ids.ServiceId;
import com.unclouded.android.logging.Logging;
import com.unclouded.android.remoteref.Letter;
import com.unclouded.android.remoteref.RemoteReferencesThreadPool;
import com.unclouded.android.services.DataService;
import com.unclouded.android.services.MultiServiceSubscription;
import com.unclouded.android.services.ObjectService;
import com.unclouded.android.services.RemoteObjectService;
import com.unclouded.android.services.RemoteService;
import com.unclouded.android.services.RemoteServiceListener;
import com.unclouded.android.services.Service;
import com.unclouded.android.services.ServiceManager;
import com.unclouded.android.services.SingleServiceSubscription;

@SuppressWarnings("serial")
public class InternalEventLoop extends EventLoop {
	
	private static InternalEventLoop instance;
	
	private EventLoopId eventLoopId;
	
	public final RemoteReferencesThreadPool remoteReferencesThreadPool;
	
	private final LinkedHashMap<PromiseId, InternalPromise> promises;
	
	/** the communication bus for this actor */
	private final CommunicationBus communicationBus;
	
	/** Managing all publications and subscriptions */
	private final ServiceManager serviceManager;
	
	private InternalNetwork network;
	
	private final ActorAddressBook addressBook;
	
	
	/** -----------------------------------------------
	 * 	GET INSTANCE
	 *  ----------------------------------------------- */
	
	static public InternalEventLoop getInstance(){
		if(instance == null){
			instance = new InternalEventLoop();
		}
		return instance;
	}
	
	
	
	/** -----------------------------------------------
	 * 	PRIVATE CONSTRUCTOR
	 *  ----------------------------------------------- */
	
	private InternalEventLoop(){
		super();
		super.start();
		eventLoopId = new EventLoopId();
		communicationBus = new CommunicationBus(this);
		addressBook = new ActorAddressBook();
		remoteReferencesThreadPool = new RemoteReferencesThreadPool(this);
		promises = new LinkedHashMap<PromiseId, InternalPromise>();
		serviceManager = new ServiceManager();
	}	
	
	
	
	/** -----------------------------------------------
	 * 	GET ACTOR ID
	 *  ----------------------------------------------- */
	
	public EventLoopId getId() {
		return eventLoopId;
	}
	
	
	/** -----------------------------------------------
	 * 	REGISTER PROMISES
	 *  ----------------------------------------------- */
	
	public void registerAsyncInvocation(RemoteObjectService service, PromiseId id, InternalPromise promise){
		remoteReferencesThreadPool.event_serve(service);
		promises.put(id, promise);
	}
		
	
	
	/** -----------------------------------------------
	 * 
	 * 	LOCAL EVENTS
	 * 
	 *  ----------------------------------------------- */


	public Network goOnline() {
		if(network == null){ network = new InternalNetwork(); }
		receive(new Event("goOnline") {
			public void process(Object myself) {
				if(!communicationBus.isOnline()){ // Only go online when currently offline
					Address myAddress = communicationBus.connect();
					Logging.Actor_LOG.debug("I'm online with address "+ myAddress);
					network.triggerIsOnline(myAddress.getIpAddress());
				}
			}
		});
		return (Network) network;
	}
	
	// Explicitly notify all peers of disconnection
	public void goOffline() {
		receive(new Event("goOffline") {
			public void process(Object myself) {
				Logging.Actor_LOG.debug("I'm disconnected now");
				Address address = communicationBus.getNetworkAddress();
				communicationBus.disconnect();
				network.triggerIsOffline(address.getIpAddress());
			}
		});	
	}
	
	
	public void reset(){
		receive(new Event("Reset") {
			public void process(Object myself) {
				communicationBus.disconnect();
				addressBook.reset();
				promises.clear();
				serviceManager.reset();
				if(network != null){
					network.cancel();
					network = null;
				}
			}
		});
	}
	
	
	/* -----------------------------
	 * BROADCAST
	 * ----------------------------- */
	
	private ServicePublication _broadcast(final TypeTag typeTag, final Service service) {
		final InternalServicePublication publication = new InternalServicePublication(typeTag, service, this);
		receive(new Event("Publish Service"){
			public void process(Object myself) {
				serviceManager.registerPublication(publication);
				if(communicationBus.isOnline()){ // Export only if actor is online
					publication.publish(communicationBus);
				} 
			}
		});
		return (ServicePublication) publication;
	}
	
	public ServicePublication broadcast(final TypeTag typeTag, final UObject object) {
		final Service service = new ObjectService(object, this);
		return _broadcast(typeTag, service);
	}
	
	
	public ServicePublication broadcast(final TypeTag typeTag, final String string){
		final Service service = new DataService(string, this);
		return _broadcast(typeTag, service);
	}
	
	public ServicePublication broadcast(final TypeTag typeTag, final Integer integer){
		final Service service = new DataService(integer, this);
		return _broadcast(typeTag, service);
	}
	
	public ServicePublication broadcast(final TypeTag typeTag, final Boolean bool){
		final Service service = new DataService(bool, this);
		return _broadcast(typeTag, service);
	}
	
	public ServicePublication broadcast(final TypeTag typeTag, final Serializable serializable){
		final Service service = new DataService(serializable, this);
		return _broadcast(typeTag, service);
	}
	
	
	
	/* -----------------------------
	 * DISCOVERY
	 * ----------------------------- */
	
	public ServiceSubscription when(final TypeTag typeTag, final ServiceListener<? extends Object> listener){
		RemoteServiceListener<?> listenerObj = new RemoteServiceListener<Object>(listener);
		final SingleServiceSubscription subscription = new SingleServiceSubscription(typeTag, listenerObj, this);
		receive(new Event("Discover Service") {
			public void process(Object myself) {
				serviceManager.registerSubscription(subscription, typeTag);
				Logging.Actor_LOG.debug("Service subscription "+ subscription +" registered");
				if(communicationBus.isOnline()){ // Only look for remote services when actor is online
					Collection<RemoteService> services = serviceManager.getRemoteServices(typeTag);
					for(RemoteService service : services){
						boolean hasJoined = subscription.serviceJoined(service);
						if(hasJoined){ break; }
					}
				}
			}
		});
		return (ServiceSubscription) subscription;
	}
	
	
	public ServiceSubscription whenever(final TypeTag typeTag, final ServiceListener<? extends Object> listener){
		RemoteServiceListener<?> listenerObj = new RemoteServiceListener<Object>(listener);
		final MultiServiceSubscription subscription = new MultiServiceSubscription(typeTag, listenerObj, this);
		receive(new Event("Discover Service") {
			public void process(Object myself) {
				serviceManager.registerSubscription(subscription, typeTag);
				Logging.Actor_LOG.debug("Service subscription "+ subscription +" registered");
				if(communicationBus.isOnline()){ // Only look for remote services when actor is online
					Collection<RemoteService> services = serviceManager.getRemoteServices(typeTag);
					for(RemoteService service : services){
						subscription.serviceJoined(service);
					}
				}
			}
		});
		return (ServiceSubscription) subscription;
	}
	
	
	public void cancelServiceSubscription(final TypeTag typeTag, final InternalServiceSubscription subscription){
		receive(new Event("Cancel Subscription") {
			public void process(Object myself) {
				serviceManager.cancelSubscription(typeTag, subscription);
			}
		});	
	}
	
	protected void cancelServicePublication(final PublicationId publicationId){
		receive(new Event("Cancel Publication") {
			public void process(Object myself) {
				InternalServicePublication publication = serviceManager.cancelPublication(publicationId);
				if(publication != null){
					TypeTag typeTag = publication.getTypeTag();
					ServiceId serviceId = publication.getService().getServiceId();
					new CMDCancelService(typeTag, serviceId).send(communicationBus);
				}
			}
		});			
	}
	
	public void sendLetter(final Letter letter, final EventLoopId replyToActorId){
		receive(new Event("Send Letter") {
			public void process(Object myself) {
				Address replyToAddress = addressBook.getAddressOf(replyToActorId);
				if(replyToAddress != null){
					new CMDTransmitLetter(letter).send(communicationBus, replyToAddress);
				}
			}
		});
	}

	
	
	
	/** -----------------------------------------------
	 * 
	 * 	NETWORK EVENTS
	 * 
	 *  ----------------------------------------------- */
	
	public void event_memberJoined(final Address address, final EventLoopId eventLoopId) {
		receive(new Event("memberJoined("+address+")") {
			public void process(Object myself) {
				Logging.Actor_LOG.debug("Discovered new actor: " + address);
				addressBook.addEntry(eventLoopId, address);
				// Export all buffered publications
				for(InternalServicePublication publication : serviceManager.getAllPublications()){
					publication.publish(communicationBus);
				}
				network.triggerIsConnectedTo(address.getIpAddress());
			}
		});
	}
	
	
	public void event_memberLeft(final Address address) {
		receive(new Event("memberLeft("+address+")") {
			public void process(Object myself) {
				
				Logging.Actor_LOG.debug("Disconnection of actor: " + address);
				EventLoopId remoteActorId = addressBook.getActorIdOf(address);
				addressBook.removeEntry(address);
				
				Collection<RemoteService> disconnectedServices = serviceManager.getRemoteServices(remoteActorId);
				for(RemoteService service : disconnectedServices){
					service.hasDisconnected();
				}
				
				network.triggerIsDisconnectedFrom(address.getIpAddress());
			}
		});
	}
	
	public void event_objectServiceJoined(final TypeTag typeTag, final ServiceId serviceId) {
		receive(new Event("Service Discovered") {
			public void process(Object myActorMirror) {
				
				RemoteService service = null;
				if(serviceManager.isNewService(serviceId)){
					Logging.Actor_LOG.debug("New service "+typeTag+" joined");
					service = serviceManager.registerRemoteObjectService(typeTag, serviceId);
				} else {
					Logging.Actor_LOG.debug("Already existing service "+typeTag+" joined");
					service = serviceManager.getRemoteService(serviceId);
				}
				Collection<InternalServiceSubscription> subscriptions = serviceManager.getSubscriptions(typeTag);
				for(InternalServiceSubscription subscription : subscriptions){
					subscription.serviceJoined(service);
				}
				
			}
		});
	}
	
	public void event_dataServiceJoined(final TypeTag typeTag, final ServiceId serviceId, final Serializable data) {
		receive(new Event("Service Discovered") {
			public void process(Object myActorMirror) {
				
				RemoteService service = null;
				if(serviceManager.isNewService(serviceId)){
					Logging.Actor_LOG.debug("New service "+typeTag+" joined");
					service = serviceManager.registerRemoteDataService(typeTag, serviceId, data);
				} else {
					Logging.Actor_LOG.debug("Already existing service "+typeTag+" joined");
					service = serviceManager.getRemoteService(serviceId);
				}
				Collection<InternalServiceSubscription> subscriptions = serviceManager.getSubscriptions(typeTag);
				for(InternalServiceSubscription subscription : subscriptions){
					subscription.serviceJoined(service);
				}
				
			}
		});
	}


	
	@SuppressWarnings("rawtypes")
	public void event_letterReceived(final Letter letter, final Address replyTo){
		receive(new Event("Letter Received") {
			public void process(Object myActorMirror) {
				ServiceId serviceId = letter.getReceivingServiceId();
				ObjectService service = (ObjectService) serviceManager.getService(serviceId);
				UObject obj = service.getObject();
				if (obj != null){
					String selector = letter.getSelector();
					Serializable[] args = letter.getArguments();
					Class[] params = new Class[args.length];
					for(int i = 0; i < params.length; i++){
						params[i] = args[i].getClass();
					}
					try {
						Method method = obj.getClass().getMethod(selector, params);
						Object[] objArgs = (Object[]) args;
						Serializable result = (Serializable) method.invoke(obj, objArgs);
						new CMDTransmitLetterReply(letter, result).send(communicationBus, replyTo);
					} catch (NoSuchMethodException e) {
						e.printStackTrace();
					} catch (IllegalArgumentException e) {
						e.printStackTrace();
					} catch (IllegalAccessException e) {
						e.printStackTrace();
					} catch (InvocationTargetException e) {
						e.printStackTrace();
					}
				}				
			}
		});
	}
	
	
	public void event_letterReplyReceived(final PromiseId promiseId, final Serializable answer) {
		receive(new Event("Letter Reply Received") {
			public void process(Object myActorMirror) {
				InternalPromise promise = promises.get(promiseId);
				if(promise != null){ promise.resolve(answer); }
			}
		});
	}
	
	
	public void event_serviceCanceled(final TypeTag typeTag, final ServiceId serviceId){
		receive(new Event("Remote Service Canceled") {
			public void process(Object object) {
				RemoteService service = serviceManager.getRemoteService(serviceId);
				service.hasDisconnected();
			}
		});
	}
	

}
