package com.unclouded.android.services;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;

import com.unclouded.android.TypeTag;
import com.unclouded.android.ids.EventLoopId;
import com.unclouded.android.ids.PublicationId;
import com.unclouded.android.ids.ServiceId;
import com.unclouded.android.internal.InternalServicePublication;
import com.unclouded.android.internal.InternalServiceSubscription;
import com.unclouded.android.util.MultiMap;

public class ServiceManager {

	/** A list of publication objects that represent locally exported service objects. */
	private final HashMap<PublicationId, InternalServicePublication> publications;
	
	/** A list of all services */
	private final HashMap<ServiceId, Service> services;	
	
	/** private final MultiMap<TypeTag, UncloudedRemoteService> remoteServicesPerTypeTag; */
	private final MultiMap<EventLoopId, RemoteService> remoteServicesPerActorId;
	private final MultiMap<TypeTag, RemoteService> remoteServicesPerTypeTag;
	private final HashMap<ServiceId, RemoteService> remoteServicesPerId;
	
	/** A list of Subscription objects that represent local subscription handlers. */
	private final MultiMap<TypeTag, InternalServiceSubscription> subscriptions;
	
	public ServiceManager(){
		subscriptions = new MultiMap<TypeTag, InternalServiceSubscription>();
		publications = new HashMap<PublicationId, InternalServicePublication>();
		services = new HashMap<ServiceId, Service>();
		
		remoteServicesPerId = new HashMap<ServiceId, RemoteService>();
		remoteServicesPerActorId = new MultiMap<EventLoopId, RemoteService>();
		remoteServicesPerTypeTag = new MultiMap<TypeTag, RemoteService>();
		
	}
	
	
	/* ----------------------------
	 * PUBLICATIONS
	 * ---------------------------- */
	
	public void registerPublication(InternalServicePublication publication){
		PublicationId publicationId = publication.getId();
		Service service = publication.getService();
		ServiceId serviceId = service.getServiceId();
		services.put(serviceId, service);
		publications.put(publicationId, publication);
	}
	
	public InternalServicePublication cancelPublication(PublicationId id){
		InternalServicePublication publication = publications.remove(id);
		if(publication != null){
			ServiceId serviceId = publication.getService().getServiceId();
			services.remove(serviceId);
		}
		return publication;
	}
	
	public Collection<InternalServicePublication> getAllPublications(){
		return publications.values();
	}
	
	public Service getService(ServiceId id){
		return services.get(id);
	}
	
	
	/* ----------------------------
	 * SUBSCRIPTIONS
	 * ---------------------------- */
	
	public void registerSubscription(InternalServiceSubscription subscription, TypeTag typeTag){
		subscriptions.add(typeTag, subscription);		
	}
	
	public void cancelSubscription(TypeTag typeTag, InternalServiceSubscription subscription){
		subscriptions.removeValue(typeTag, subscription);
	}
	
	public Collection<InternalServiceSubscription> getSubscriptions(TypeTag typeTag){
		return subscriptions.getValues(typeTag);
	}

	
	/* ----------------------------
	 * REMOTE SERVICES
	 * ---------------------------- */
	
	public boolean isNewService(ServiceId serviceId){
		return !(remoteServicesPerId.containsKey(serviceId));
	}
	
	public RemoteService registerRemoteObjectService(TypeTag typeTag, ServiceId serviceId){
		RemoteService newService = new RemoteObjectService(serviceId);
		registerRemoteService(typeTag, newService, serviceId);
		return newService;
	}
	
	
	public RemoteService registerRemoteDataService(TypeTag typeTag, ServiceId serviceId, Serializable data){
		RemoteService newService = new RemoteDataService(serviceId, data);
		registerRemoteService(typeTag, newService, serviceId);
		return newService;
	}
	
	private void registerRemoteService(TypeTag typeTag, RemoteService service, ServiceId serviceId){
		EventLoopId actorId = serviceId.getActorId();
		remoteServicesPerId.put(serviceId, service);
		remoteServicesPerActorId.add(actorId, service);
		remoteServicesPerTypeTag.add(typeTag, service);
	}
	
	
	public Collection<RemoteService> getRemoteServices(EventLoopId actorId){
		return remoteServicesPerActorId.getValues(actorId);
	}
	
	public Collection<RemoteService> getRemoteServices(TypeTag typeTag){
		return remoteServicesPerTypeTag.getValues(typeTag);
	}
	
	public RemoteService getRemoteService(ServiceId id){
		return remoteServicesPerId.get(id);
	}
	
	
	
	
	/* ---------------------------- */
	
	public void reset(){
		publications.clear();
		services.clear();
		remoteServicesPerActorId.clear();
		remoteServicesPerTypeTag.clear();
		remoteServicesPerId.clear();
		subscriptions.clear();
	}
	
	

	

}
