/**
 * AmbientTalk/2 Project
 * MultiMap.java
 * (c) Programming Technology Lab, 2006 - 2007
 * Authors: Tom Van Cutsem & Stijn Mostinckx
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */
package com.unclouded.android.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * A simple Map allowing one key to be associated with a Set of values.
 * 
 * @author smostinc & ndecooma
 */
public class MultiMap<K, V> extends HashMap<K, Set<V>> {


	private static final long serialVersionUID = -2937418667767489134L;
	
	public boolean containsValue(Object value) {
		Collection<Set<V>> valuesSets = super.values();
		for (Iterator<Set<V>> iter = valuesSets.iterator(); iter.hasNext();) {
			Set<V> element = (Set<V>) iter.next();
			if(element.contains(value))
				return true;
		}
		return false;
	}

	public Set<Map.Entry<K, V>> getEntrySet() {
		Set<Map.Entry<K, V>> result = new HashSet<Map.Entry<K, V>>();
		Set<K> keys = super.keySet();
		for (Iterator<K> iter = keys.iterator(); iter.hasNext();) {
			K key = iter.next();
			Set<V> values = (Set<V>)get(key);
			for (Iterator<V> iterator = values.iterator(); iterator.hasNext();) {
				V value = iterator.next();
				result.add(new MyEntry<K, V>(key, value));
			}
		}
		return result;
	}
	
	public Set<V> getValues(K key){
		Set<V> values = super.get(key);
		if(values == null){
			values = new HashSet<V>();
		}
		return values;
	}

	
	public V add(K key, V value) {
		Set<V> existingValues = super.get(key);
		if(existingValues == null) { 
			Set<V> values = new HashSet<V>();
			values.add(value); 
			super.put(key, values);
		} else {
			existingValues.add(value);
		}
		return value;
	}

	public void addAll(K key, Collection<V> values) {
		Set<V> existingValues = super.get(key);
		if(existingValues == null) {
			Set<V> setOfValues = new HashSet<V>();
			setOfValues.addAll(values); 
			super.put(key, setOfValues);
		} else {
			existingValues.addAll(values);
		}
	}


	/**
	 * Removes a single value associated with a key 
	 */
	public boolean removeValue(K key, V value) {
		Set<V> values = (Set<V>)super.get(key);
		if(values != null) {
			return values.remove(value);
		} else {
			return false;
		}
	}

	public Collection<V> getAllValues() {
		Collection<V> result = new ArrayList<V>();
		Collection<Set<V>> values = super.values();
		for (Iterator<Set<V>> iter = values.iterator(); iter.hasNext();) {
			Set<V> set = iter.next();
			for (Iterator<V> iterator = set.iterator(); iterator.hasNext();) {
				V value = iterator.next();
				result.add(value);
			}
		}
		return result;
	}

	
	@SuppressWarnings("unused")
	private void replaceValue(K key, V oldValue, V newValue) {
		Set<V> values = (Set<V>)super.get(key);
		values.remove(oldValue);
		values.add(newValue);
	}
	
	
	// An alternative to defining this class is: new AbstractMap.SimpleEntry<K, V>(key, value);
	// This requires Java API v9 however
	final class MyEntry<Key, Value> implements Map.Entry<Key, Value> {
	    private final Key key;
	    private Value value;

	    public MyEntry(Key key, Value value) {
	        this.key = key;
	        this.value = value;
	    }

	    @Override
	    public Key getKey() {
	        return key;
	    }

	    @Override
	    public Value getValue() {
	        return value;
	    }

	    @Override
	    public Value setValue(Value value) {
	        Value old = this.value;
	        this.value = value;
	        return old;
	    }
	}


}


